#ifndef __resummation__
#define __resummation__

#include <string>
#include "luminosity.hh"
#include "math/complex.hh"

using namespace std;




namespace resumm {

  // actual alphas routine will be provided by the user
  extern double alphas(double mu);

  enum log_order{
    LL    = 0,
    NLL   = 1,
    NNLL  = 2,
    NNNLL = 3
  };
  enum order{
    LO    = 0,
    NLO   = 1,
    NNLO  = 2,
    NNNLO = 3
  };
  enum prescription {
    BP00 =  10,  // resums (log^k log(1/z) / log(1/z))_+
    BP0  =  11,  // resums (log^k log(1/z) / log(1/z))_+ /sqrt(z)
    BP1  =  1,   // resums (log^k (1-z) / (1-z))_+
    BP2  =  2,   // resums (log^k ((1-z)/sqrt(z)) / (1-z))_+
    //
    MP00 = -10,  // same logs as BP00 (equivalent in practice): log(N)
    MP0  = -11,  // same logs as BP0  (equivalent in practice): log(N-1/2)
    MP2  = -2    // using psi0(N) instead of log(N-1/2) (formally almost equivalent to BP2, and similar to MP0=BP0)
  };
  enum APtype{
    AP0  = 0,  // AP to order 0
    AP1  = 1,  // AP to order 1
    AP2  = 2,  // AP to order 2
    AP12 = 12, // average between AP1 and AP2
    AP   = -2  // full AP
  };
  enum LogCountType{
    primed = 0,
    unstarred = 0,
    appropriateforC = 0,
    Catani = 0,
    //
    unprimed = 1,
    starred = 1,
    appropriateforlogC = 1,
    Neubert = 1
  };

  extern string sPrescr(prescription p);
  extern string sAP(APtype ap);

  enum verbosity { SILENT=0, LOWKEY=1, DEFAULT=2 };

  class Resummation {

  public:
    //Resummation(process p, string PDFset, int PDFsubset=0, verbosity v=DEFAULT);
    Resummation(process p, collider c, string PDFset, int PDFsubset=0, verbosity v=DEFAULT);
    ~Resummation();

    // re-initialize PDFs
    void PDFinit(string PDFset, int PDFsubset=0, verbosity v=DEFAULT);
    void init_mass_thresholds(double mc, double mb, double mt);
    void Setmuf(double mu); // I should delete this (used just for luminosity tests)
    void SetMtop   (double mt); // value used in Higgs resummation (in principle not the same as top threshold)
    void SetMbottom(double mb);
    void SetMcharm (double mc);
    void UseLargeMtopEFT(bool use=true); // set to false for the exact result
    //friend double Lum(double x, Resummation res);
    friend double Lum(double x, void *p);

    // select prescription (which mainly means "select subleading terms")
    // last argument is effective only with Borel summation
    void SetPrescriptionByName(string pr, bool exponentiate_g0);
    void SetPrescription(prescription p=BP2, bool exponentiate_g0=true, bool resumDistrConst=false, bool expBorelPrescription=false);
    void SetAPimprovement(APtype impr=AP1);
    void SetBPpars(unsigned int Ncheb=10, double cutoff=2.);
    void SetMPpars(unsigned int Ncheb=10);
    void SetMPpars(unsigned int Ncheb, double beta, int gamma, double z_min=1.e-5);
    void SetLogCounting(LogCountType lct = primed);
    void ComputeCommon(bool redo);

    double xsLL   (double m, double sqrt_s, double muF_m_ratio=1., double muR_m_ratio=1., order match_order=LO  , double *asinput=NULL);
    double xsNLL  (double m, double sqrt_s, double muF_m_ratio=1., double muR_m_ratio=1., order match_order=NLO , double *asinput=NULL);
    double xsNNLL (double m, double sqrt_s, double muF_m_ratio=1., double muR_m_ratio=1., order match_order=NNLO, double *asinput=NULL);
    double xsNNNLL(double m, double sqrt_s, double muF_m_ratio=1., double muR_m_ratio=1., order match_order=NNLO, double *asinput=NULL);
    virtual double res_xs (double m, double sqrt_s, double muF_m_ratio, double muR_m_ratio, log_order LOGo, order match_order, double *asinput=NULL) =0;

    double xsRapLL   (double m, double Y, double sqrt_s, double muF_m_ratio=1., double muR_m_ratio=1., order match_order=LO  , double *asinput=NULL);
    double xsRapNLL  (double m, double Y, double sqrt_s, double muF_m_ratio=1., double muR_m_ratio=1., order match_order=NLO , double *asinput=NULL);
    double xsRapNNLL (double m, double Y, double sqrt_s, double muF_m_ratio=1., double muR_m_ratio=1., order match_order=NNLO, double *asinput=NULL);
    double xsRapNNNLL(double m, double Y, double sqrt_s, double muF_m_ratio=1., double muR_m_ratio=1., order match_order=NNLO, double *asinput=NULL);
    virtual double res_xs_rap(double m, double Y, double sqrt_s, double muF_m_ratio, double muR_m_ratio, log_order LOGo, order match_order, double *asinput=NULL) =0;

    double  g0();
    double g01, g02, g03;

  protected:
    double mass_c, mass_b, mass_t; // used as thresholds
    double as_c, as_b, as_t;
    //
    process  proc;
    collider coll;
    bool doRapidityDistr;
    prescription prescr;
    APtype APimpr;
    bool g0exp;
    bool expDconst;
    bool expBP;
    //
    unsigned int Nchebyshev;
    double *coeff;
    //
    double BPcutoff;
    double MPbeta, MPzmin;
    int MPgamma;
    double MP_reN0, MP_slope;
    double DefaultBeta();
    int DefaultGamma();
    //
    log_order LL_order;
    LogCountType LogarithmicCounting;
    order matching_order;
    //
    double tau, luminosity_in_tau;
    double rapY;
    double M, muF, muR, mtop, mbot, mcha;
    int nf;
    int nf_of_mu(double mu);
    double as, ab;
    double b0, b1, b2, b3;
    double A1, A2, A3, A4, D2, D3;
    double d1c, d2c, d3c;
    double gA0, gA1, gB0, gB1, gB2;
    //double g01, g02, g03;
    //double d1, d2, d3;
    double g0sub1, g0sub2, g0sub3;
    double g0sub1s, g0sub2s, g0sub3s;
    double g0bar1, g0bar2, g0bar3;
    double W1, W2, W3;
    bool mtEFT;
    //
    double Nlandau();
    //
    dcomplex mellD(int, dcomplex);
    dcomplex APimprovement(dcomplex (Resummation::*f)(dcomplex,void*), dcomplex N, void *p=NULL);
    //double  g0();
    double  g0sub();
    dcomplex g1(dcomplex);
    dcomplex g2(dcomplex);
    dcomplex g3(dcomplex);
    dcomplex g4(dcomplex);
    dcomplex S(dcomplex, void*);
    dcomplex asS1 (dcomplex,void*);
    dcomplex as2S2(dcomplex,void*);
    dcomplex as3S3(dcomplex,void*);
    double asg01 ();
    double as2g02();
    double as3g03();
    dcomplex g0expS_order1(dcomplex);
    dcomplex g0expS_order2(dcomplex);
    dcomplex g0expS_order3(dcomplex);
    dcomplex expS(dcomplex);
    dcomplex g0expS_matched(dcomplex);
    //
    void InitResummCoeffs();
    void InitCoeffDY();
    void InitConstDY();
    void InitCoeffHiggs();
    void InitConstHiggs();
    void InitConstHiggsCommons();
    //
    dcomplex delta_term(dcomplex);
    dcomplex cheb_term(dcomplex, int);
    dcomplex convolution(dcomplex);
    dcomplex BorelIntegrand(double, dcomplex);
    double  BorelIntegrand_chvar(double, double);
    friend int integrandCUBA(const int *ndim, const double x[], const int *ncomp, double res[], void *pars);
    double BorelPrescription();
    //
    dcomplex NLogGen(dcomplex, void*);
    dcomplex BorelIntegrandExp(double, dcomplex, dcomplex);
    dcomplex BorelIntegrandExp_chvar(double, double, dcomplex);
    friend int integrandCUBAexp(const int *ndim, const double x[], const int *ncomp, double res[], void *pars);
    dcomplex BorelPrescriptionExp(dcomplex);
    //
    dcomplex LuminosityN(dcomplex N);
    dcomplex lum_times_g0expS_matched(dcomplex N);
    friend double Minimal_integrand(double t, void *p);
    friend double Minimal_integrand_inf(double t, void *p);
    dcomplex InverseMellinIntegrand(dcomplex N);
    double InverseMellin();
    double MinimalPrescription();
    //
    // cache
    double last_muFrat;
    double last_muRrat;
    double last_M;
    double last_sqrts;
    double last_mtop;
    int last_nf;
    int last_Ncheb;
    bool last_EFT;
    bool firstinit;


  };


  double Lum(double x, void *p);

};




#endif
